import time
from functools import wraps
from .context import contextdecorator
from . import iters

Exc = Exception


@contextdecorator
def ignore(exc=Exc):
    '''Ignore exception raised.'''
    try:
        yield
    except exc:
        pass

def throttle_exception(seconds, exc=Exc):
    '''Throttle a function if an exception was raised. Useful for retrying on
    failure.'''
    def outer(func):
        def inner(*a, **kw):
            try:
                t0 = time.time()
                return func(*a, **kw)
            except exc as e:
                time.sleep(max(0, seconds - (time.time() - t0)))
                raise e
        return inner
    return outer


def retry_on_failure(n=5, exc=Exception):
    def outer(func):
        class X: pass
        def inner(*a, **kw):
            X.e = Exception('Retried {} times.'.format(n))
            for _ in iters.limit(iters.infinite(), n):
                try:
                    return func(*a, **kw)
                except exc as e:
                    X.e = e
                    continue
            raise X.e
        return inner
    return outer


def default_value(value, exc=Exception):
    '''Return a default value if an exception was raised.'''
    def outer(func):
        @wraps(func)
        def inner(*a, **kw):
            try:
                return func(*a, **kw)
            except exc:
                return value() if callable(value) else value
        return inner
    return outer



from functools import wraps
import contextlib


def contextdecorator(func):
    '''Like contextlib.contextmanager, but the wrapped function also
    doubles as a decorator.

    Example:

        @contextdecorator
        def blah(a, b): # classdecorator.inner
            print(a)
            yield
            print(b)

        @blah(4, 5) # ContextDecorator.__init__
        def xyz(): # ContextDecorator.__call__
            print(1)
        xyz() # ContextDecorator.__call__.inner
        # prints 4, 1, 5

        with blah(4, 5): # ContextDecorator.__init__, ContextDecorator.__enter__
            print(1)
        # prints 4, 1, 5
    '''
    @wraps(func)
    def inner(*a, **kw):
        cm = _ContextDecorator(func, *a, **kw)
        cm.wrapper = inner.wrapper
        return cm
    inner.wrap = setter(inner, 'wrapper')
    return inner


def setter(obj, attr, default=None):
    def inner(value):
        setattr(obj, attr, value)
    inner(default)
    return inner


class _ContextDecorator(contextlib._GeneratorContextManager):
    '''Helper for @contextdecorator decorator.'''
    wrapper = None
    def __init__(self, func, *a, **kw):
        self.func, self.a, self.kw = func, a, kw
        self.__doc__ = getattr(func, "__doc__", None) or type(self).__doc__

    def __enter__(self):
        self.gen = self.func(*self.a, **self.kw)
        try:
            return next(self.gen)
        except StopIteration:
            raise RuntimeError("generator didn't yield") from None

    def __call__(self, func):
        @wraps(func)
        def inner(*a, **kw):
            with self:
                return func(*a, **kw)
        return self.wrapper(inner) if callable(self.wrapper) else inner




EMPTY = object()

class cachedproperty:
    """A property that is only computed once per instance and then replaces itself
    with an ordinary attribute. Deleting the attribute resets the property."""
    def __init__(self, func):
        self.__doc__ = getattr(func, '__doc__')
        self.func = func
        self.cacheid = '_cached~{}'.format(self.func.__name__)

    def __get__(self, obj, cls):
        d = obj.__dict__
        if self.cacheid not in d:
            d[self.cacheid] = self.func(obj)
        return d[self.cacheid]

class onceproperty:
    def __init__(self, func):
        self.__doc__ = getattr(func, '__doc__')
        self.func = func
        self.value = EMPTY

    def __get__(self, obj, cls):
        if self.value is EMPTY:
            self.value = self.func(obj)
        return self.value

class overridable_property:
    def __init__(self, func):
        self.name = getattr(func, '__name__', None) or 'func{}'.format(id(func))
        self.__doc__ = (
            getattr(func, '__doc__', None) or
            'Overridable Property: {}'.format(self.name))
        self.value_name = '_{}'.format(self.name)
        self.func = func
        self.unset = None

    def __get__(self, obj, cls):
        value = getattr(obj, self.value_name, self.unset)
        return self.func(obj) if value is self.unset else value

    def __set__(self, obj, value):
        setattr(obj, self.value_name, value)



import time
from functools import wraps


class StopScheduler(Exception):
    '''A special exception that will stop the scheduler.'''

class _IgnoreScheduler(Exception): # dummy exception - no one would throw this
    '''Default ignored exception. Something no one would actually throw.'''


def schedule_task(freq, priority=1, initial_delay=0, ignore=None, stop_signal=StopScheduler, sc=None):
    '''After a task has finished running, schedule it again unless a
    contradicting signal is received.
    '''
    import sched
    sc = sc or sched.scheduler(time.time, time.sleep)
    freq_init = freq if initial_delay is None else initial_delay
    ignore = ignore or _IgnoreScheduler

    def wrapper(func):
        schedule = lambda args, kwargs, freq=freq: sc.enter(
            freq, priority, inner, args, kwargs)

        @wraps(func)
        def inner(*a, **kw):
            try: # try to run the function
                _ = func(*a, **kw)
                if _ == stop_signal: # possibly use return value to exit
                    return

                # update arguments using return result
                if isinstance(_, tuple):
                    a = _
                elif isinstance(_, dict):
                    kw = _
            except StopScheduler: # exit if the built-in stop is raised
                return
            except ignore: # custom ignore exceptions
                pass

            schedule(a, kw)

        def run(*a, **kw):
            schedule(a, kw, freq=freq_init)
            sc.run()
        return run
    return wrapper



import time
import itertools
from . import misc


def limit(it, n=None):
    return it if n is None else (x for i, x in zip(range(n), it))


def asfunc(it):
    it = iter(it)
    return lambda: next(it)


def infinite(i=0, inc=1):
    '''Infinite generator. Turn a while loop into a for loop.'''
    while True:
        yield i
        i += inc


def throttled(seconds=1):
    '''Force an iterable to take at minimum x seconds.'''
    dt = asleep = 0
    while True:
        t0 = time.time()
        yield dt, asleep # time asleep
        if seconds:
            dt = time.time() - t0
            asleep = max(seconds - dt, 0)
            time.sleep(asleep)


def pre_check_iter(it, n=1):
    '''Check the value first n items of an iterator without unloading them
    from the iterator queue.'''
    it = iter(it)
    items = [_ for i, _ in zip(range(n), it)]
    return items, itertools.chain(items, it)


def run_iter_forever(get_iter, none_if_empty=None, throttle=None, timeout=None):
    '''Return a never ending iterable.

    Arguments:
        get_iter (func): creates a new iterator.
        none_if_empty (bool, optional): if True, return None on an empty iterator.
        as_func (bool): whether to return an iterator or it.__next__
        throttle (float): throttle to make sure iterations aren't faster than
            `throttle` seconds.
        timeout (float): if no iterations are produced for x seconds, return None.
    '''
    required_item = int(bool(none_if_empty or timeout))
    def forever():
        t0 = timeout and time.time()
        for _ in throttled(throttle):
            # get the iterator and see if it has at least n items
            items, it = pre_check_iter(get_iter() or (), required_item)

            # check if iterable is considered empty and should yield None
            if len(items) >= required_item:
                # yield iterator items
                t0 = timeout and time.time()
                yield from it
            elif none_if_empty or timeout and time.time() - t0 > timeout:
                yield None
    return forever()



import sys
import time
from functools import wraps
from . import misc
from .context import contextdecorator


def log_error_as_warning(logger, *a, **kw):
    '''Instead of throwing an error, log as a warning.'''
    return handle_error(logger.warning, *a, **kw)

def log_error_as_error(logger, *a, **kw):
    '''Instead of throwing an error, log as a warning.'''
    return handle_error(logger.error, *a, **kw)

def ignore_error(ignore=(Exception,), **kw):
    '''Ignore error thrown.'''
    return handle_error(None, exc=(), ignore=ignore, **kw)


def handle_error(log, msg='', default=None, **kw):
    '''Catch errors in function call

    Arguments:
        log (callable): the logging function (e.g. logger.warning).
        msg (str): The logging message to display on failure.
            (e.g. 'Upload failed:' will print 'Upload failed: (TypeError) this is an error message.')
        exc (Exception, tuple of Exceptions): the exceptions to catch.
        default (any): the value to return if an exception is raised.

    '''
    def outer(func):
        msg_ = msg or 'Exception in {}: '.format(func.__qualname__)
        return misc.default_value(default)(
            catch_log_error(log, msg_, **kw)(func))
    return outer

@contextdecorator
def catch_log_error(log, msg, should_raise=False, exc=(Exception,),
                    log_traceback=False, ignore=(), exit=False,
                    throttle=None):
    '''Catch errors thrown within context manager.

    Arguments:
        should_raise (bool): should we still raise the exception? default: False.
        log_traceback (bool): if True, log the traceback as well.
        ignore (Exception, tuple of Exceptions): any exceptions to ignore.
        exit (int): if specified, call sys.exit(exit) where exit is the exit code.
        throttle (float): if specified, sleep to ensure that the function takes
            `throttle` seconds if an error is thrown. This is to prevent
            functions going into rapid failure cycles.
    '''
    try:
        yield
    except ignore:
        pass
    except exc as e:
        # TODO: any way to pass exception and use that as a trace? idk this is FINE for now
        #       later reading: https://stackoverflow.com/questions/1278705/when-i-catch-an-exception-how-do-i-get-the-type-file-and-line-number
        log(msg + ' ({}) {}'.format(type(e).__name__, e),
            extra={'override': get_exec_info(drop=1)},
            exc_info=log_traceback)
        if should_raise:
            raise e
        if exit or exit is 0: # int or True
            exit_code = int(1 if exit is True else exit)
            log('Exiting with code: %d', exit_code)
            sys.exit(exit_code)


@catch_log_error.wrap
def _logcallwrap(func):
    pass


def log_exec_time(logger, msg='{name} took {sec}s'):
    '''Log the function execution time.'''
    def outer(func):
        @wraps(func)
        def inner(*a, **kw):
            t0 = time.time()
            try:
                return func(*a, **kw)
            finally:
                t = time.time() - t0
                logger.info(msg.format(
                    sec=t, ms=t*1000, min=t/60.,
                    name=func.__name__))
        return inner
    return outer



###############
# Utils
###############


def gather_tbs():
    tbs = [sys.exc_info()[-1]]
    while True:
        if tbs[-1].tb_next is None:
            return tbs[::-1]
        tbs.append(tbs[-1].tb_next)

def get_exec_info(depth=0, sep='<', drop=0):
    exc_tbs = gather_tbs()
    exc_tbs = exc_tbs[depth:len(exc_tbs) - drop]
    return {
        'filename': exc_tbs[0].tb_frame.f_code.co_filename,
        'funcName': sep.join(tb.tb_frame.f_code.co_name for tb in exc_tbs),
        'lineno'  : exc_tbs[0].tb_lineno,
    }



import functools
import itertools
import collections
import inspect
from inspect import Parameter as Pm


POSARG_TYPES = [Pm.POSITIONAL_ONLY, Pm.POSITIONAL_OR_KEYWORD]
ARG_TYPES = POSARG_TYPES + [Pm.KEYWORD_ONLY]

Spec = collections.namedtuple('Spec', 'args, vararg, varkw, posargs, kwargs, defaults')

def get_argspec(f):
    '''Get function parameters'''
    ps = inspect.signature(f).parameters.items()
    args = [n for n, p in ps if p.kind in ARG_TYPES]
    defaults = {n: p.default for n, p in ps if p.kind in ARG_TYPES}
    posargs = [n for n, p in ps if p.kind in POSARG_TYPES and
               p.default == inspect._empty]
    vararg = next((n for n, p in ps if p.kind == Pm.VAR_POSITIONAL), None)
    varkw = next((n for n, p in ps if p.kind == Pm.VAR_KEYWORD), None)
    kwargs = {n: p.default for n, p in ps if p.default != inspect._empty}
    return Spec(args, vararg, varkw, posargs, kwargs, defaults)


def filterkw(func):
    '''Remove arguments that aren't in the function signature.'''
    spec = get_argspec(func)
    if spec.varkw:
        return func

    args = set(spec.args)
    @functools.wraps(func)
    def inner(*a, **kw):
        return func(*a, **{k: v for k, v in kw.items() if k in args})
    return inner


class configfunction:
    '''Allow functions to have easily overrideable default arguments.
    Works with mixed positional and keyword arguments.

    @configfunction
    def abc(a=5, b=6):
        return a + b
    assert abc() == 11
    abc.update(a=10)
    assert abc() == 16
    assert abc(2) == 8

    '''
    def __init__(self, func, fill_varkw=True, view=None):
        self.function = func
        (self.spec) = get_argspec(func)
        functools.update_wrapper(self, func)

        self.config = {}
        self.fill_varkw = fill_varkw

    def __call__(self, *a, _cfg=None, **kw):
        a, kw = self.config_args(self.merge_config(_cfg), *a, **kw)
        return self.function(*a, **kw)

    def config_args(self, cfg, *a, **kw):
        # fill in any positional arguments we can
        a += tuple(cfg[x] for x in itertools.takewhile(
            lambda x: x not in kw and x in cfg,
            self.spec.posargs[len(a):]))

        # update keywords with vars not passed in call
        kw.update((k, cfg[k]) for k in (
            set(cfg) - set(self.spec.posargs[:len(a)]) - set(kw)
            if self.fill_varkw and self.spec.varkw else
            set(cfg) & set(self.spec.args[len(a):]) - set(kw)))
        return a, kw

    def merge_config(self, cfg):
        return self.config if cfg is None else dict(self.config, **cfg)

    def update(self, *a, **kw):
        self.config.update(*a, **kw)
        return self

    def clear(self, *a):
        self.config = (
            {k: v for k, v in self.config.items() if k not in a} if a else {})
        return self
